/** BGM Direct Draw Library (by Fwirt)
 *  This file contains functions for performing direct draw on
 *  a series of chars contained within a BGMap, as an alternative
 *  to direct framebuffer drawing.
 */
#ifndef _BGMDRAW_H_
#define _BGMDRAW_H_

#define BGMAP_WIDTH 64 // The width of a BGMap

// Get hword for point x,y in a BGMap w chars wide
#define HwordNum(w,x,y) ((((x/8) * 8) + (((y/8) * 8) * w)) + (y%8))
// Get the number of the half-nybble in the hword for point x
#define ShiftNum(x) (x%8)*2 

/** Initialize BGMap b, with enough Char references to fill an
 *  area w chars wide and h chars tall starting at Char c, and
 *  clear out all used chars.
 */
void bgm_set(u8 b, u8 w, u8 h, u16 c) {
    u16* bgm_ptr = (u16*)(BGMap(b));
    u8 i, j, k;
    for (i=0; i<h; i++) // Rows
        for (j=0; j<w; j++) { // Columns
            bgm_ptr[j + (i * BGMAP_WIDTH)] = c + j + (i * w);
            for (k=0; k<8; k++) // Clear out the char being assigned
                CM[(c*8)+(i*w*8)+(j*8)+k] = 0;
        }
}

/** Set point x, y to value i, in an area w Chars wide which
 *  starts referencing at Char c.
 */
void point_set(u16 x, u16 y, u8 i, u8 w, u16 c) {
    CM[HwordNum(w,x,y) + (c*8)] &= ~(3 << ShiftNum(x));
    CM[HwordNum(w,x,y) + (c*8)] |= (i & 3) << ShiftNum(x);
}

#endif
