/** Mandelbrot Explorer for VB Copyright (c) 2010, Collin "Fwirt" Skiles **/

#include "libgccvb/libgccvb.h"
#include "tiles.h"
#include "map.h"
#include "chadjustment.h"
#include "bgadjustment.h"
#include "bgmdraw.h"
#include "sgasket.h"

#define SEGSIZE 8192 // Size of one BGMap/Character segment
#define BWIDTH 48 // Width of area of BGMap in tiles
#define BHEIGHT 28 // Height
#define BCSTART 80 // Character to start grabbing from
#define SWIDTH 384 // Screen width
#define SHEIGHT 224 // Height
#define RLSTART -2.809357 // Initial zoom location
#define RHSTART 1.476357
#define ILSTART -1.25
#define IHSTART 1.25
#define CMAXSTART 50 // Max iterations to determine if a point is in the set

// Prototypes
void text_out(u8,u8,u8,char*);
u8 colorize(float,float);
void render(void);
void startup_screens(void);
void ovr_set_num(float,float);
void ovr_set_steps(u16);

// Global variables
float r_min = RLSTART;
float r_max = RHSTART;
float i_min = ILSTART;
float i_max = IHSTART;
u16 count_max = CMAXSTART;
u16* const BGMLOC = (void*)BGMap(0); // The location of the first BGMap

int main() {
    vertex cursor = {(SWIDTH / 2) - 3, (SHEIGHT / 2) - 3}; // Vertex is already defined in sgasket so we might as well use it
    vertex box;
    u16 status_count = 0;
    u16 gamepad;
    u8 a_down = 0;
    u8 cursor_show = 1;

    vbDisplayOn();
    startup_screens(); // Show startup screens

    // Render world
    bgm_set(1, BWIDTH, BHEIGHT, BCSTART);
    vbSetWorld(31, WRLD_ON | WRLD_BGMAP | WRLD_BGMBASE(1), 0, 0, 0, 0, 0, 0, SWIDTH-1, SHEIGHT-1);
    // Status bar and cursor world
    VIP_REGS[JPLT0] = 0xC0;
    VIP_REGS[JPLT1] = 0x04;
    VIP_REGS[SPT3] = 84;
    VIP_REGS[SPT2] = 0;
    for (u8 i=1; i<81; i++) { // Status bar objects
        OBJ_XSET(i, 295+i);
        OBJ_PSET(i, -2);
        OBJ_YSET(i, SHEIGHT-16);
        OBJ_CSET(i, 78);
        OBJ_VIS(i, 3);
    }
    for (u8 i=84; i>80; i--) { // Cursor objects
        OBJ_XSET(i, cursor.x);
        OBJ_PSET(i, -1);
        OBJ_YSET(i, cursor.y);
        OBJ_CSET(i, 40);
        OBJ_VIS(i, 3);
    }
    vbSetWorld(30, WRLD_ON | WRLD_OBJ, 0, 0, 0, 0, 0, 0, SWIDTH-1, SHEIGHT-1);
    // Overlay world
    VIP_REGS[GPLT1] = 0xD0;
    vbSetWorld(29, WRLD_ON | WRLD_BGMAP | WRLD_BGMBASE(0), 0, -2, 0, 0, 0, 0, SWIDTH-1, SHEIGHT-1);
    WA[28].head = WRLD_END; // End world

    // Fill in the overlay data    
    ovr_set_num(r_min + (cursor.x + 3) * (r_max - r_min) / SWIDTH, i_min + (cursor.y + 3) * (i_max - i_min) / SHEIGHT); // Update overlay
    ovr_set_steps(count_max);

    vbDisplayShow();

    i_max = i_min + SHEIGHT * (r_max - r_min) / SWIDTH; // Set i according to aspect ratio
    render();

    // Setup the loop timer
    timer_freq(TIMER_100US);
    timer_set(TIME_MS(15));
    timer_clearstat();
    timer_enable(1);


    for(;;) {
        // Key held cases
        gamepad = vbReadPad();
        if ((gamepad & K_LPAD) && cursor_show) {
            if (a_down) { // Update the zoom frame
                box.y = cursor.y + SHEIGHT * (box.x - cursor.x) / SWIDTH; // Equalize aspect ratio
                if ((gamepad & (K_LL | K_LU)) && (box.x+3 > 0) && (box.x > cursor.x))
                    box.x--;
                else if ((gamepad & (K_LR | K_LD)) && (box.x+4 < SWIDTH-2))
                    box.x++;
                for (u8 i=84; i>80; i--) { // Update cursor
                    OBJ_CSET(i, 41);
                    OBJ_HSET(i, i%2);
                    OBJ_VSET(i, ((i-80)/2)%2);
                    OBJ_XSET(i, cursor.x + (i % 2 * (box.x - cursor.x)));
                    OBJ_YSET(i, cursor.y + (((i-80)/2)%2 * (box.y - cursor.y)));
                }
                // ovr_set_num(r_min + (box.x + 3) * (r_max - r_min) / SWIDTH, i_min + (box.y + 3) * (i_max - i_min) / SHEIGHT); // Update overlay
            }
            else { // Update the cursor
                if ((gamepad & K_LU) && (cursor.y+3 > 0))
                    cursor.y--;
                else if ((gamepad & K_LD) && (cursor.y+3 < SHEIGHT-1))
                    cursor.y++;
                if ((gamepad & K_LL) && (cursor.x+3 > 0))
                    cursor.x--;
                else if ((gamepad & K_LR) && (cursor.x+3 < SWIDTH-2))
                    cursor.x++;
                for (u8 i=84; i>80; i--) { // Update cursor
                    OBJ_XSET(i, cursor.x);
                    OBJ_YSET(i, cursor.y);
                }
                ovr_set_num(r_min + (cursor.x + 3) * (r_max - r_min) / SWIDTH, i_min + (cursor.y + 3) * (i_max - i_min) / SHEIGHT); // Update overlay
            }
        }

        gamepad = vbPadKeyDown(); // Key pressed cases
        // Select Button
        if (gamepad & K_SEL) { // Hide the cursor
            cursor_show ^= 1;
            for (u8 i=84; i>80; i--)
                OA[i].jp ^= 0xC000;
        }
        // Start Button
        if (gamepad & K_STA) { // Reset zoom
            r_min = RLSTART;
            r_max = RHSTART;
            i_min = ILSTART;
            i_max = i_min + SHEIGHT * (r_max - r_min) / SWIDTH; // Set i according to aspect ratio
            ovr_set_num(r_min + (cursor.x + 3) * (r_max - r_min) / SWIDTH, i_min + (cursor.y + 3) * (i_max - i_min) / SHEIGHT); // Update overlay
            render();
        }    
        // Left trigger
        if (gamepad & K_LT) { // Toggle the overlay
            WA[29].head ^= WRLD_ON;
            for (u8 i=1; i<81; i++)
                OA[i].jp ^= 0xC000;
        }
        // Right trigger
        if (gamepad & K_RT) // Refresh the screen
            render();
        // A Button
        if (gamepad & K_A) { // Go into zoom mode
            a_down = 1;
            box.x = cursor.x;
            box.y = cursor.y;
        }
        // B Button
        if (gamepad & K_B) { // Zoom out by a factor of 2
            float width = (r_max - r_min);
            r_min -= width / 2;
            r_max += width / 2;
            i_min -= (i_max - i_min) / 2;
            i_max = i_min + SHEIGHT * (r_max - r_min) / SWIDTH;
            ovr_set_num(r_min + (cursor.x + 3) * (r_max - r_min) / SWIDTH, i_min + (cursor.y + 3) * (i_max - i_min) / SHEIGHT); // Update overlay
            render();
        }
        // Right pad
        if (gamepad & K_RPAD) { // Adjust calculation steps
            if ((gamepad & K_RU) && (count_max < 9900))
                count_max += 100;
            else if ((gamepad & K_RD) && (count_max > 100))
                count_max -= 100;
            if ((gamepad & K_RR) && (count_max < 9990))
                count_max += 10;
            else if ((gamepad & K_RL) && (count_max > 10))
                count_max -= 10;
            ovr_set_steps(count_max);
        }

        gamepad = vbPadKeyUp(); // Key released cases

        // A Button
        if (gamepad & K_A) { // Zoom to the selected frame
            float prev;
            a_down = 0;
            for (u8 i=84; i>80; i--) { // Update cursor
                OBJ_CSET(i, 40);
                OBJ_HSET(i, 0);
                OBJ_VSET(i, 0);
                OBJ_XSET(i, cursor.x);
                OBJ_YSET(i, cursor.y);
            }
            if (box.x != cursor.x) {
                prev = r_max;
                r_max = r_min + (box.x + 3) * (r_max - r_min) / SWIDTH;
                r_min = r_min + (cursor.x + 3) * (prev - r_min) / SWIDTH;
                i_min = i_min + (cursor.y + 3) * (i_max - i_min) / SHEIGHT;
                i_max = i_min + SHEIGHT * (r_max - r_min) / SWIDTH; // Scale preserving aspect ratio
                ovr_set_num(r_min + (cursor.x + 3) * (r_max - r_min) / SWIDTH, i_min + (cursor.y + 3) * (i_max - i_min) / SHEIGHT); // Update overlay
                render();
            }
        }

        // Wait for the loop timer
        while (!timer_getstat());
        timer_clearstat();

        // Occasionally change the statusbar back to normal
        status_count++;
        if (status_count > 100) {
            // Reset bar to black
            for (u8 o=1; o<81; o++)
                OBJ_PALSET(o, 0);
            // Change status to "Mandelbrot Explorer"
            for (u16 b=0; b<10; b++)
                BGMLOC[1701+b] = BGMLOC[49+b];
            status_count = 0;
        }
    }
    return 0;
}


/** Render the view window **/
void render(void) {
    float tr = (r_max - r_min) / SWIDTH;
    float ti = (i_max - i_min) / SHEIGHT;
    // Reset bar to black
    for (u8 o=1; o<81; o++)
        OBJ_PALSET(o, 0);
    // Change status to "RENDERING"
    for (u16 b=0; b<10; b++)
        BGMLOC[1701+b] = BGMLOC[113+b];
    for (u16 i=0; i<SHEIGHT; i++)
        for (u16 j=0; j<SWIDTH; j++) {
            // Draw a point in the set
            point_set(j, i, colorize(r_min + j * tr, i_min + i * ti), BWIDTH, BCSTART);
            OBJ_PALSET((i*80)/(SHEIGHT-1),1); // Fill the progress bar
        }
    // Change status to "DONE"
    for (u16 b=0; b<10; b++)
        BGMLOC[1701+b] = BGMLOC[177+b];
}

/** Return a color depending on how many iterations
 *  it takes for the input point to escape the set
 */
u8 colorize(float r, float i) {
    u16 c = 0;
    float zr = 0.0f;
    float zi = 0.0f;
    float zi2;
    float total = 0.0f;

    // Calculate the number of passes until this point escapes the set
    do {
        zi2 = zi * zi;
        total = zr * zr + zi2;
        zi = 2.0f * zr * zi + i;
        zr = zr * zr - zi2 + r;
        c++;
    } while ((c < count_max) && (total < 4.0f));

    // Return a color ramp of the result
    if (c == count_max)
        return 0;
    else if (c > (count_max / 2))
        return 1;
    else if (c > (count_max / 4))
        return 2;
    else
        return 3;
}

/** Shows the warning and calibration screens **/
void startup_screens() {
    // Load the font and screen maps into memory
    copymem((u8*) CharSeg0, (u8*) TILES, SEGSIZE);
    copymem((u8*) BGMap(0), (u8*) MAP, SEGSIZE);
    copymem((u8*) CharSeg2, (u8*) CHADJUSTMENT, SEGSIZE);
    copymem((u8*) BGMap(1), (u8*) BGADJUSTMENT, SEGSIZE);

    // Setup Sierpinski Gasket
    sg_setup(200, 173, 2, 25, 22, BCSTART, 584931);

    // Setup and display caution screen
    vbSetWorld(31, WRLD_ON | WRLD_BGMAP | WRLD_BGMBASE(2), 91, 3, 26, 0, 0, 0, 200, 173);
    //vbSetWorld(30, WRLD_ON | WRLD_BGMAP, 112, 0, 76, 0, 0, 232, 159, 56);
    vbSetWorld(30, WRLD_ON | WRLD_BGMAP, 16, 0, 16, 0, 0, 232, 159, 56);
    WA[29].head = WRLD_END;
    
    // Nintendo specs require 2 seconds before the caution screen goes away :P
    timer_freq(TIMER_100US);
    timer_set(TIME_SEC(2));
    timer_clearstat();
    timer_enable(1);

    vbDisplayShow(0);
    while(!timer_getstat() || !(vbReadPad() & K_ANY)) {
        sg_next_point(1);
        // A little delay
        for (u8 count = 0; count < 250; count++);
    }
    vbFXFadeOut(0);

    timer_clearstat();
    timer_enable(0);

    // Setup and display calibration screen
    vbSetWorld(31, WRLD_LON | WRLD_BGMBASE(1), 0, 0, 0, 0, 0, 0, SWIDTH, SHEIGHT);
    vbSetWorld(30, WRLD_RON | WRLD_BGMBASE(1), 0, 0, 0, 0, 0, 224, SWIDTH, SHEIGHT);
    WA[29].head = WRLD_END;

    vbFXFadeIn(0);
    while(!(vbReadPad() & K_ANY));
    vbFXFadeOut(0);
}

/** Puts the complex number defined by r+i into the overlay boxes **/
void ovr_set_num(float r, float i) {
    int a = (int)(r*10000000);
    int b = (int)(i*10000000);
    // Setup sign and make sure we're positive
    if (a < 0) {a *= -1; BGMLOC[130] = (0x4000 | 13);}
    else BGMLOC[130] = (0x4000 | 48);
    if (b < 0) {b *= -1; BGMLOC[143] = (0x4000 | 13);}
    else BGMLOC[143] = (0x4000 | 48);

    for (u8 c=10; c>3; c--) {
        BGMLOC[129+c] = (0x4000 | ((a%10)+1));
        BGMLOC[142+c] = (0x4000 | ((b%10)+1));
        a /= 10;
        b /= 10;
    }
    BGMLOC[131] = (0x4000 | ((a%10)+1));
    BGMLOC[144] = (0x4000 | ((b%10)+1));
}

/** Puts the number i into the "STEPS" box in the overlay **/
void ovr_set_steps(u16 i) {
    for (u8 c=4; c>0; c--) {
        BGMLOC[155+c] = (((i%10)+1) | 0x4000);
        i /= 10;
    }        
}
