/** Sierpinski for Virtual Boy (c) 2010, Fwirt **/
/** This is a library derived from a chaos game demo
 *  that I wrote. It contains functions for drawing
 *  a Sierpinski gasket to a BGMap. (Using bgmdraw.h)
 */
#ifndef _SGASKET_H_
#define _SGASKET_H_

#include "libgccvb/libgccvb.h"
#include "bgmdraw.h"

typedef struct vertex {
    s16 x;
    s16 y;
} vertex;

//Global variables
static vertex tri_vertex[3];
static vertex point;
static u16 bgm_width;
static u16 ch_num;

/** Finds a vertex between a and b and stores the result in b **/
void sg_between(vertex* a, vertex* b) {
    vertex vector = {b->x - a->x, b->y - a->y}; // Vector from a to b
    vector.x /= 2; // Halve the magnitude of the vector
    vector.y /= 2;
    b->x = a->x + vector.x;
    b->y = a->y + vector.y; // Add the vector to a to get the point
}

/** Add another point in our Sierpinski gasket **/
void sg_next_point(u8 color) {
    // Get the point between i and one of the vertices
    sg_between(&tri_vertex[random()%3],&point);

    // Draw the point to the correct char in our BGMap
    point_set(point.x, point.y, color, bgm_width, ch_num);
}

void sg_setup(u16 triwidth, u16 triheight, u8 bgm, u8 bgmw, u8 bgmh, u8 ch, u32 seed) {
    // Seed the random
    seed_random(seed);

    // Setup static variables
    bgm_width = bgmw;
    ch_num = ch;
    
    // Set up the BGMap for drawing
    bgm_set(bgm, bgm_width, bgmh, ch);

    // Setup the triangle vertices
    tri_vertex[0].x = triwidth/2;
    tri_vertex[1].y = triheight;
    tri_vertex[2].x = triwidth;
    tri_vertex[2].y = triheight;

    vbDisplayShow();

    // Get a point inside the triangle
    point.x = triwidth/2;
    point.y = triheight/2;
}

#endif
